//+------------------------------------------------------------------+
//|                                                 indstats.mq{4|5} |
//|                                      Copyright © 2017, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//|           Simple expert adviser for testing probability approach |
//|                            https://www.mql5.com/ru/articles/3264 |
//+------------------------------------------------------------------+
#property copyright "Copyright © 2017, Marketeer"
#property link      "https://www.mql5.com/en/users/marketeer"
#property version   "1.0"
#property description "Probability driven trading system, based on Bayes rules and indicators' statistics"

#property strict

input int ConsistentSignalNumber = 1;
input bool PrintDetails = false;
input int _Magic = 0; // Magic
input float Lot = 0.01f;
input int TradeDuration = 1;

#ifdef __MQL5__
  #include <IndStats/MarketMQL4.mqh>
  #include <IndStats/ind4to5.mqh>
  #define MT4ORDERS_FASTHISTORY_OFF
  #include <mt4orders.mqh> // https://www.mql5.com/ru/code/16006
#endif

#include <IndStats/IndicatN.mqh>
#include <IndStats/Expert0.mqh>
#include <IndStats/HashMapTemplate.mqh>


int OnInit()
{
  return IndicatN::handleInit();
}

// statistics to be used for probabilities estimation
int bars = 0; // total count of bars/samples
int bull = 0, bear = 0; // number of bars/samples per trade type
int buy[MAX_SIGNAL_NUM] = {0}, sell[MAX_SIGNAL_NUM] = {0};  // unconditional signals arrays
int buyOnBull[MAX_SIGNAL_NUM] = {0}, sellOnBear[MAX_SIGNAL_NUM] = {0}; // conditional (successful) signals arrays

HashMapTemplate<string,int> mapOfGoodStats("n/a", 0);
HashMapTemplate<string,int> mapOfAllStats("n/a", 0);

bool tradeDetected = false;
int skipped = 0;


Expert e(_Magic, Lot);

void OnTick()
{
  static datetime lastBar;
  
  if(lastBar != Time[0])
  {
    const RubbArray<#ifdef __MQL5__ IndicatN:: #endif TradeSignals> *ts = IndicatN::handleStart();
    
    bool up = false, down = false;
    int buySignalCount = 0, sellSignalCount = 0;
    
    for(int i = 0; i < ts.size(); i++)
    {
      string id = IndicatN::GetSignal(ts[i].index);
      // alerts are used to collect statistics
      if(ts[i].alert)
      {
        // while setting up events, enumerated by i,
        // hypothesis H_xxx should come first, before signals S_xxx,
        // because we assign up or down marks here
        if(id == "H_BULL")
        {
          bull++;
          buy[ts[i].index]++;
          up = true;
        }
        else if(id == "H_BEAR")
        {
          bear++;
          sell[ts[i].index]++;
          down = true;
        }
        else if(StringFind(id, "S_BUY") == 0)
        {
          buy[ts[i].index]++;
          mapOfAllStats.inc(StringSubstr(id, 5), 1);
          if(up)
          {
            if(PrintDetails) Print("buyOk ", id);
            buyOnBull[ts[i].index]++;
            mapOfGoodStats.inc(StringSubstr(id, 5), 1);
          }
        }
        else if(StringFind(id, "S_SELL") == 0)
        {
          sell[ts[i].index]++;
          mapOfAllStats.inc(StringSubstr(id, 6), 1);
          if(down)
          {
            if(PrintDetails) Print("sellOk ", id);
            sellOnBear[ts[i].index]++;
            mapOfGoodStats.inc(StringSubstr(id, 6), 1);
          }
        }
        
        if(PrintDetails) Print(ts[i].message);
      }
      else
      if(ts[i].buy)
      {
        if(PrintDetails) Print("buy ", id, " ", DoubleToString(ts[i].value, _Digits));
        buySignalCount++;
        tradeDetected = true;
      }
      else
      if(ts[i].sell)
      {
        if(PrintDetails) Print("sell ", id, " ", DoubleToString(ts[i].value, _Digits));
        sellSignalCount++;
        tradeDetected = true;
      }
    }

    if(e.getLastOrderBar() >= TradeDuration)
    {
      e.closeMarketOrders();
    }
    
    if(buySignalCount >= ConsistentSignalNumber
    && sellSignalCount >= ConsistentSignalNumber)
    {
      Print("Signal collision");
    }
    else
    if(buySignalCount >= ConsistentSignalNumber)
    {
      e.closeMarketOrders(e.mask(OP_SELL));
      
      if(e.getOrderCount(e.mask(OP_BUY)) == 0)
      {
        e.placeMarketOrder(OP_BUY);
      }
      else
      {
        Print("Skipped buy");
        skipped++;
      }
    }
    else
    if(sellSignalCount >= ConsistentSignalNumber)
    {
      e.closeMarketOrders(e.mask(OP_BUY));
      
      if(e.getOrderCount(e.mask(OP_SELL)) == 0)
      {
        e.placeMarketOrder(OP_SELL);
      }
      else
      {
        Print("Skipped sell");
        skipped++;
      }
    }
    
    lastBar = Time[0];
    bars++;
  }
}

void OnDeinit(const int reason)
{
  if(reason != REASON_PROGRAM /*MT4*/ && reason != REASON_REMOVE /*MT5*/) return;
  
  Print("bars=", bars);
  if(!tradeDetected) // if no trades, assume statistics collection
  {
    Print("bull=", bull, " bear=", bear);
    Array1DPrint("   buy", buy, MAX_SIGNAL_NUM, 0, 6);
    Array1DPrint(" buyOk", buyOnBull, MAX_SIGNAL_NUM, 0, 6);
    Array1DPrint("  sell", sell, MAX_SIGNAL_NUM, 0, 6);
    Array1DPrint("sellOk", sellOnBear, MAX_SIGNAL_NUM, 0, 6);
    
    double totals[MAX_SIGNAL_NUM];
    ArrayInitialize(totals, 0);
    for(int i = 0; i < MAX_SIGNAL_NUM; i++)
    {
      if(buy[i] + sell[i] > 0)
      {
        totals[i] = (double)(buyOnBull[i] + sellOnBear[i]) / (buy[i] + sell[i]);
      }
    }
    Array1DPrint("totals", totals, MAX_SIGNAL_NUM, 2, 6);
    double probs[];
    ArrayResize(probs, mapOfAllStats.getSize());
    Print("Stats by name:");
    for(int i = 0; i < mapOfAllStats.getSize(); i++)
    {
      int good = mapOfGoodStats[mapOfAllStats.getKey(i)];
      int all = mapOfAllStats[i];
      double prob = (double)good / all;
      probs[i] = prob;
      Print(mapOfAllStats.getKey(i), "=", DoubleToString(prob, 3), " [", good, "/", all, "]");
    }
    
    double pn = 1, mn = 1;
    for(int i = 0; i < mapOfAllStats.getSize(); i++)
    {
      pn *= probs[i];
      mn *= 1 - probs[i];
    }
    
    Print("Overall probablity estimate: ", DoubleToString(pn / (pn + mn), 3));
  }
  else // if trades occured, estimate actual performance
  {
    int buyOk = 0, sellOk = 0;
    int buyCount = 0, sellCount = 0;
    
    for(int i = 0; i < OrdersHistoryTotal(); i++)
    {
      if(OrderSelect(i, SELECT_BY_POS, MODE_HISTORY))
      {
        int b = iBarShift(NULL, 0, OrderOpenTime());
        int t = MathMax(b - TradeDuration, 0);
        if(OrderType() == OP_BUY)
        {
          buyCount++;
          if(Open[t] - Open[b] > 0)
          {
            buyOk++;
          }
          if(PrintDetails) Print("#", OrderTicket(), " ", TimeToString(Time[b]), " ", (int)((Open[t] - Open[b]) / Point));
        }
        else
        if(OrderType() == OP_SELL)
        {
          sellCount++;
          if(Open[b] - Open[t] > 0)
          {
            sellOk++;
          }
          if(PrintDetails) Print("#", OrderTicket(), " ", TimeToString(Time[b]), " ", (int)((Open[b] - Open[t]) / Point));
        }
      }
    }
    
    Print("Buys: ", buyOk, "/", buyCount, " ", buyCount>0?DoubleToString((double)buyOk/buyCount, 2):"", " Sells: ",
          sellOk, "/", sellCount, " ", sellCount>0?DoubleToString((double)sellOk/sellCount, 2):"", " Total: ",
          (buyCount + sellCount)>0?DoubleToString((double)(buyOk + sellOk) / (buyCount + sellCount), 2):"");
    if(skipped > 0)
    {
      Print("Trades skipped:", skipped);
    }
  }
}

template<typename T>
void Array1DPrint(string title, T &v[], int dx, int precision = 0, int width = 0)
{
  string s = "";
  for(int i = 0; i < dx; i++)
  {
    string a = DoubleToString(v[i], precision);
    if(StringLen(a) < width)
    {
      string padding;
      StringInit(padding, width - StringLen(a), ' ');
      a = padding + a;
    }
    s = s + a + " ";
  }
  Print(title + ": " + s);
}
